{# form, form_data_key, form_schema_key, and form_ui_schema_key are passed in #}
{{ form.get_data_for_template|json_script:form_data_key }}
{{ form.get_data_schema|json_script:form_schema_key }}
{{ form.get_ui_schema|json_script:form_ui_schema_key }}
{{ form.get_errors_for_template|json_script:form_errors_key }}
{{ form.get_custom_elements_content|json_script:custom_elements_key }}